@@ -29,7 +29,6 @@ import (
 	"golang.org/x/tools/internal/imports"
 	"golang.org/x/tools/internal/lsp/protocol"
 	"golang.org/x/tools/internal/lsp/source"
-	"golang.org/x/tools/internal/memoize"
 	"golang.org/x/tools/internal/span"
 	"golang.org/x/tools/internal/xcontext"
 	errors "golang.org/x/xerrors"
@@ -93,10 +92,6 @@ type View struct {
 	// workspaceInformation tracks various details about this view's
 	// environment variables, go version, and use of modules.
 	workspaceInformation
-
-	// tempWorkspace is a temporary directory dedicated to holding the latest
-	// version of the workspace go.mod file. (TODO: also go.sum file)
-	tempWorkspace span.URI
 }
 
 type workspaceInformation struct {
@@ -141,16 +136,8 @@ const (
 
 	// tempModfile indicates whether or not the -modfile flag should be used.
 	tempModfile
-
-	// usesWorkspaceModule indicates support for the experimental workspace module
-	// feature.
-	usesWorkspaceModule
 )
 
-type builtinPackageHandle struct {
-	handle *memoize.Handle
-}
-
 // fileBase holds the common functionality for all files.
 // It is intended to be embedded in the file implementations
 type fileBase struct {
@@ -231,10 +218,6 @@ func (v *View) Folder() span.URI {
 	return v.folder
 }
 
-func (v *View) TempWorkspace() span.URI {
-	return v.tempWorkspace
-}
-
 func (v *View) Options() *source.Options {
 	v.optionsMu.Lock()
 	defer v.optionsMu.Unlock()
@@ -255,6 +238,8 @@ func (v *View) FileKind(fh source.FileHandle) source.FileKind {
 		return source.Mod
 	case ".sum":
 		return source.Sum
+	case ".work":
+		return source.Work
 	}
 	exts := v.Options().TemplateExtensions
 	for _, ext := range exts {
@@ -396,11 +381,10 @@ func (s *snapshot) locateTemplateFiles(ctx context.Context) {
 		excluded := pathExcludedByFilter(relpath, dir, s.view.gomodcache, s.view.options)
 		if fileHasExtension(path, suffixes) && !excluded && !fi.IsDir() {
 			k := span.URIFromPath(path)
-			fh, err := s.GetVersionedFile(ctx, k)
+			_, err := s.GetVersionedFile(ctx, k)
 			if err != nil {
 				return nil
 			}
-			s.files[k] = fh
 		}
 		searched++
 		if fileLimit > 0 && searched > fileLimit {
@@ -735,63 +719,12 @@ func (v *View) invalidateContent(ctx context.Context, changes map[span.URI]*file
 
 	oldSnapshot := v.snapshot
 
-	var workspaceChanged bool
-	v.snapshot, workspaceChanged = oldSnapshot.clone(ctx, v.baseCtx, changes, forceReloadMetadata)
-	if workspaceChanged {
-		if err := v.updateWorkspaceLocked(ctx); err != nil {
-			event.Error(ctx, "copying workspace dir", err)
-		}
-	}
+	v.snapshot = oldSnapshot.clone(ctx, v.baseCtx, changes, forceReloadMetadata)
 	go oldSnapshot.generation.Destroy("View.invalidateContent")
 
 	return v.snapshot, v.snapshot.generation.Acquire()
 }
 
-func (v *View) updateWorkspace(ctx context.Context) error {
-	if v.tempWorkspace == "" {
-		return nil
-	}
-	v.snapshotMu.Lock()
-	defer v.snapshotMu.Unlock()
-	return v.updateWorkspaceLocked(ctx)
-}
-
-// updateWorkspaceLocked should only be called when v.snapshotMu is held. It
-// guarantees that workspace module content will be copied to v.tempWorkace at
-// some point in the future. We do not guarantee that the temp workspace sees
-// all changes to the workspace module, only that it is eventually consistent
-// with the workspace module of the latest snapshot.
-func (v *View) updateWorkspaceLocked(ctx context.Context) error {
-	if v.snapshot == nil {
-		return errors.New("view is shutting down")
-	}
-
-	release := v.snapshot.generation.Acquire()
-	defer release()
-	src, err := v.snapshot.getWorkspaceDir(ctx)
-	if err != nil {
-		return err
-	}
-	for _, name := range []string{"go.mod", "go.sum"} {
-		srcname := filepath.Join(src.Filename(), name)
-		srcf, err := os.Open(srcname)
-		if err != nil {
-			return errors.Errorf("opening snapshot %s: %w", name, err)
-		}
-		defer srcf.Close()
-		dstname := filepath.Join(v.tempWorkspace.Filename(), name)
-		dstf, err := os.Create(dstname)
-		if err != nil {
-			return errors.Errorf("truncating view %s: %w", name, err)
-		}
-		defer dstf.Close()
-		if _, err := io.Copy(dstf, srcf); err != nil {
-			return errors.Errorf("copying %s: %w", name, err)
-		}
-	}
-	return nil
-}
-
 func (s *Session) getWorkspaceInformation(ctx context.Context, folder span.URI, options *source.Options) (*workspaceInformation, error) {
 	if err := checkPathCase(folder.Filename()); err != nil {
 		return nil, errors.Errorf("invalid workspace folder path: %w; check that the casing of the configured workspace folder path agrees with the casing reported by the operating system", err)
@@ -867,7 +800,7 @@ func go111moduleForVersion(go111module string, goversion int) go111module {
 // TODO (rFindley): move this to workspace.go
 // TODO (rFindley): simplify this once workspace modules are enabled by default.
 func findWorkspaceRoot(ctx context.Context, folder span.URI, fs source.FileSource, excludePath func(string) bool, experimental bool) (span.URI, error) {
-	patterns := []string{"go.mod"}
+	patterns := []string{"go.work", "go.mod"}
 	if experimental {
 		patterns = []string{"go.work", "gopls.mod", "go.mod"}
 	}
@@ -917,7 +850,8 @@ func findRootPattern(ctx context.Context, folder span.URI, basename string, fs s
 		if exists {
 			return span.URIFromPath(dir), nil
 		}
-		next, _ := filepath.Split(dir)
+		// Trailing separators must be trimmed, otherwise filepath.Split is a noop.
+		next, _ := filepath.Split(strings.TrimRight(dir, string(filepath.Separator)))
 		if next == dir {
 			break
 		}
@@ -971,6 +905,7 @@ func (s *Session) getGoEnv(ctx context.Context, folder string, goversion int, go
 	for k := range vars {
 		args = append(args, k)
 	}
+	args = append(args, "GOWORK")
 
 	inv := gocommand.Invocation{
 		Verb:       "env",
