@@ -14,16 +14,41 @@ import (
 	"go/format"
 	"go/token"
 	"go/types"
+	"strconv"
 	"strings"
 	"time"
+	"unicode/utf8"
 
+	"golang.org/x/text/unicode/runenames"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/lsp/protocol"
 	"golang.org/x/tools/internal/typeparams"
 	errors "golang.org/x/xerrors"
 )
 
-type HoverInformation struct {
+// HoverContext contains context extracted from the syntax and type information
+// of a given node, for use in various summaries (hover, autocomplete,
+// signature help).
+type HoverContext struct {
+	// signatureSource is the object or node use to derive the hover signature.
+	//
+	// It may also hold a precomputed string.
+	// TODO(rfindley): pre-compute all signatures to avoid this indirection.
+	signatureSource interface{}
+
+	// comment is the most relevant comment group associated with the hovered object.
+	Comment *ast.CommentGroup
+}
+
+// HoverJSON contains information used by hover. It is also the JSON returned
+// for the "structured" hover format
+type HoverJSON struct {
+	// Synopsis is a single sentence synopsis of the symbol's documentation.
+	Synopsis string `json:"synopsis"`
+
+	// FullDocumentation is the symbol's full documentation.
+	FullDocumentation string `json:"fullDocumentation"`
+
 	// Signature is the symbol's signature.
 	Signature string `json:"signature"`
 
@@ -31,11 +56,8 @@ type HoverInformation struct {
 	// This is recommended only for use in clients that show a single line for hover.
 	SingleLine string `json:"singleLine"`
 
-	// Synopsis is a single sentence synopsis of the symbol's documentation.
-	Synopsis string `json:"synopsis"`
-
-	// FullDocumentation is the symbol's full documentation.
-	FullDocumentation string `json:"fullDocumentation"`
+	// SymbolName is the types.Object.Name for the given symbol.
+	SymbolName string `json:"symbolName"`
 
 	// LinkPath is the pkg.go.dev link for the given symbol.
 	// For example, the "go/ast" part of "pkg.go.dev/go/ast#Node".
@@ -44,28 +66,14 @@ type HoverInformation struct {
 	// LinkAnchor is the pkg.go.dev link anchor for the given symbol.
 	// For example, the "Node" part of "pkg.go.dev/go/ast#Node".
 	LinkAnchor string `json:"linkAnchor"`
-
-	// importPath is the import path for the package containing the given
-	// symbol.
-	importPath string
-
-	// symbolName is the types.Object.Name for the given symbol.
-	symbolName string
-
-	source  interface{}
-	comment *ast.CommentGroup
-
-	// typeName contains the identifier name when the identifier is a type declaration.
-	// If it is not empty, the hover will have the prefix "type <typeName> ".
-	typeName string
-	// isTypeAlias indicates whether the identifier is a type alias declaration.
-	// If it is true, the hover will have the prefix "type <typeName> = ".
-	isTypeAlias bool
 }
 
 func Hover(ctx context.Context, snapshot Snapshot, fh FileHandle, position protocol.Position) (*protocol.Hover, error) {
 	ident, err := Identifier(ctx, snapshot, fh, position)
 	if err != nil {
+		if hover, innerErr := hoverRune(ctx, snapshot, fh, position); innerErr == nil {
+			return hover, nil
+		}
 		return nil, nil
 	}
 	h, err := HoverIdentifier(ctx, ident)
@@ -76,10 +84,6 @@ func Hover(ctx context.Context, snapshot Snapshot, fh FileHandle, position proto
 	if err != nil {
 		return nil, err
 	}
-	// See golang/go#36998: don't link to modules matching GOPRIVATE.
-	if snapshot.View().IsGoPrivatePath(h.importPath) {
-		h.LinkPath = ""
-	}
 	hover, err := FormatHover(h, snapshot.View().Options())
 	if err != nil {
 		return nil, err
@@ -93,30 +97,193 @@ func Hover(ctx context.Context, snapshot Snapshot, fh FileHandle, position proto
 	}, nil
 }
 
-func HoverIdentifier(ctx context.Context, i *IdentifierInfo) (*HoverInformation, error) {
-	ctx, done := event.Start(ctx, "source.Hover")
+func hoverRune(ctx context.Context, snapshot Snapshot, fh FileHandle, position protocol.Position) (*protocol.Hover, error) {
+	ctx, done := event.Start(ctx, "source.hoverRune")
 	defer done()
 
-	fset := i.Snapshot.FileSet()
-	h, err := HoverInfo(ctx, i.Snapshot, i.pkg, i.Declaration.obj, i.Declaration.node, i.Declaration.fullDecl)
+	r, mrng, err := findRune(ctx, snapshot, fh, position)
 	if err != nil {
 		return nil, err
 	}
+	rng, err := mrng.Range()
+	if err != nil {
+		return nil, err
+	}
+
+	var desc string
+	runeName := runenames.Name(r)
+	if len(runeName) > 0 && runeName[0] == '<' {
+		// Check if the rune looks like an HTML tag. If so, trim the surrounding <>
+		// characters to work around https://github.com/microsoft/vscode/issues/124042.
+		runeName = strings.TrimRight(runeName[1:], ">")
+	}
+	if strconv.IsPrint(r) {
+		desc = fmt.Sprintf("'%s', U+%04X, %s", string(r), uint32(r), runeName)
+	} else {
+		desc = fmt.Sprintf("U+%04X, %s", uint32(r), runeName)
+	}
+	return &protocol.Hover{
+		Contents: protocol.MarkupContent{
+			Kind:  snapshot.View().Options().PreferredContentFormat,
+			Value: desc,
+		},
+		Range: rng,
+	}, nil
+}
+
+// ErrNoRuneFound is the error returned when no rune is found at a particular position.
+var ErrNoRuneFound = errors.New("no rune found")
+
+// findRune returns rune information for a position in a file.
+func findRune(ctx context.Context, snapshot Snapshot, fh FileHandle, position protocol.Position) (rune, MappedRange, error) {
+	pkg, pgf, err := GetParsedFile(ctx, snapshot, fh, NarrowestPackage)
+	if err != nil {
+		return 0, MappedRange{}, err
+	}
+	spn, err := pgf.Mapper.PointSpan(position)
+	if err != nil {
+		return 0, MappedRange{}, err
+	}
+	rng, err := spn.Range(pgf.Mapper.Converter)
+	if err != nil {
+		return 0, MappedRange{}, err
+	}
+	pos := rng.Start
+
+	// Find the basic literal enclosing the given position, if there is one.
+	var lit *ast.BasicLit
+	var found bool
+	ast.Inspect(pgf.File, func(n ast.Node) bool {
+		if found {
+			return false
+		}
+		if n, ok := n.(*ast.BasicLit); ok && pos >= n.Pos() && pos <= n.End() {
+			lit = n
+			found = true
+		}
+		return !found
+	})
+	if !found {
+		return 0, MappedRange{}, ErrNoRuneFound
+	}
+
+	var r rune
+	var start, end token.Pos
+	switch lit.Kind {
+	case token.CHAR:
+		s, err := strconv.Unquote(lit.Value)
+		if err != nil {
+			// If the conversion fails, it's because of an invalid syntax, therefore
+			// there is no rune to be found.
+			return 0, MappedRange{}, ErrNoRuneFound
+		}
+		r, _ = utf8.DecodeRuneInString(s)
+		if r == utf8.RuneError {
+			return 0, MappedRange{}, fmt.Errorf("rune error")
+		}
+		start, end = lit.Pos(), lit.End()
+	case token.INT:
+		// It's an integer, scan only if it is a hex litteral whose bitsize in
+		// ranging from 8 to 32.
+		if !(strings.HasPrefix(lit.Value, "0x") && len(lit.Value[2:]) >= 2 && len(lit.Value[2:]) <= 8) {
+			return 0, MappedRange{}, ErrNoRuneFound
+		}
+		v, err := strconv.ParseUint(lit.Value[2:], 16, 32)
+		if err != nil {
+			return 0, MappedRange{}, err
+		}
+		r = rune(v)
+		if r == utf8.RuneError {
+			return 0, MappedRange{}, fmt.Errorf("rune error")
+		}
+		start, end = lit.Pos(), lit.End()
+	case token.STRING:
+		// It's a string, scan only if it contains a unicode escape sequence under or before the
+		// current cursor position.
+		var found bool
+		litOffset, err := Offset(pgf.Tok, lit.Pos())
+		if err != nil {
+			return 0, MappedRange{}, err
+		}
+		offset, err := Offset(pgf.Tok, pos)
+		if err != nil {
+			return 0, MappedRange{}, err
+		}
+		for i := offset - litOffset; i > 0; i-- {
+			// Start at the cursor position and search backward for the beginning of a rune escape sequence.
+			rr, _ := utf8.DecodeRuneInString(lit.Value[i:])
+			if rr == utf8.RuneError {
+				return 0, MappedRange{}, fmt.Errorf("rune error")
+			}
+			if rr == '\\' {
+				// Got the beginning, decode it.
+				var tail string
+				r, _, tail, err = strconv.UnquoteChar(lit.Value[i:], '"')
+				if err != nil {
+					// If the conversion fails, it's because of an invalid syntax, therefore is no rune to be found.
+					return 0, MappedRange{}, ErrNoRuneFound
+				}
+				// Only the rune escape sequence part of the string has to be highlighted, recompute the range.
+				runeLen := len(lit.Value) - (int(i) + len(tail))
+				start = token.Pos(int(lit.Pos()) + int(i))
+				end = token.Pos(int(start) + runeLen)
+				found = true
+				break
+			}
+		}
+		if !found {
+			// No escape sequence found
+			return 0, MappedRange{}, ErrNoRuneFound
+		}
+	default:
+		return 0, MappedRange{}, ErrNoRuneFound
+	}
+
+	mappedRange, err := posToMappedRange(snapshot, pkg, start, end)
+	if err != nil {
+		return 0, MappedRange{}, err
+	}
+	return r, mappedRange, nil
+}
+
+func HoverIdentifier(ctx context.Context, i *IdentifierInfo) (*HoverJSON, error) {
+	ctx, done := event.Start(ctx, "source.Hover")
+	defer done()
+
+	hoverCtx, err := FindHoverContext(ctx, i.Snapshot, i.pkg, i.Declaration.obj, i.Declaration.node, i.Declaration.fullDecl)
+	if err != nil {
+		return nil, err
+	}
+
+	h := &HoverJSON{
+		FullDocumentation: hoverCtx.Comment.Text(),
+		Synopsis:          doc.Synopsis(hoverCtx.Comment.Text()),
+	}
+
+	fset := i.Snapshot.FileSet()
 	// Determine the symbol's signature.
-	switch x := h.source.(type) {
+	switch x := hoverCtx.signatureSource.(type) {
+	case string:
+		h.Signature = x // a pre-computed signature
+
+	case *ast.TypeSpec:
+		x2 := *x
+		// Don't duplicate comments when formatting type specs.
+		x2.Doc = nil
+		x2.Comment = nil
+		var b strings.Builder
+		b.WriteString("type ")
+		if err := format.Node(&b, fset, &x2); err != nil {
+			return nil, err
+		}
+		h.Signature = b.String()
+
 	case ast.Node:
 		var b strings.Builder
 		if err := format.Node(&b, fset, x); err != nil {
 			return nil, err
 		}
 		h.Signature = b.String()
-		if h.typeName != "" {
-			prefix := "type " + h.typeName + " "
-			if h.isTypeAlias {
-				prefix += "= "
-			}
-			h.Signature = prefix + h.Signature
-		}
 
 		// Check if the variable is an integer whose value we can present in a more
 		// user-friendly way, i.e. `var hex = 0xe34e` becomes `var hex = 58190`
@@ -145,22 +312,7 @@ func HoverIdentifier(ctx context.Context, i *IdentifierInfo) (*HoverInformation,
 	if obj == nil {
 		return h, nil
 	}
-	switch obj := obj.(type) {
-	case *types.PkgName:
-		h.importPath = obj.Imported().Path()
-		h.LinkPath = h.importPath
-		h.symbolName = obj.Name()
-		if mod, version, ok := moduleAtVersion(h.LinkPath, i); ok {
-			h.LinkPath = strings.Replace(h.LinkPath, mod, mod+"@"+version, 1)
-		}
-		return h, nil
-	case *types.Builtin:
-		h.importPath = "builtin"
-		h.LinkPath = h.importPath
-		h.LinkAnchor = obj.Name()
-		h.symbolName = h.LinkAnchor
-		return h, nil
-	}
+
 	// Check if the identifier is test-only (and is therefore not part of a
 	// package's API). This is true if the request originated in a test package,
 	// and if the declaration is also found in the same test package.
@@ -169,66 +321,105 @@ func HoverIdentifier(ctx context.Context, i *IdentifierInfo) (*HoverInformation,
 			return h, nil
 		}
 	}
-	// Don't return links for other unexported types.
-	if !obj.Exported() {
-		return h, nil
+
+	h.SymbolName, h.LinkPath, h.LinkAnchor = linkData(obj, i.enclosing)
+
+	// See golang/go#36998: don't link to modules matching GOPRIVATE.
+	//
+	// The path returned by linkData is an import path.
+	if i.Snapshot.View().IsGoPrivatePath(h.LinkPath) {
+		h.LinkPath = ""
+	} else if mod, version, ok := moduleAtVersion(h.LinkPath, i); ok {
+		h.LinkPath = strings.Replace(h.LinkPath, mod, mod+"@"+version, 1)
 	}
-	var rTypeName string
+
+	return h, nil
+}
+
+// linkData returns the name, import path, and anchor to use in building links
+// to obj.
+//
+// If obj is not visible in documentation, the returned name will be empty.
+func linkData(obj types.Object, enclosing *types.TypeName) (name, importPath, anchor string) {
+	// Package names simply link to the package.
+	if obj, ok := obj.(*types.PkgName); ok {
+		return obj.Name(), obj.Imported().Path(), ""
+	}
+
+	// Builtins link to the special builtin package.
+	if obj.Parent() == types.Universe {
+		return obj.Name(), "builtin", obj.Name()
+	}
+
+	// In all other cases, the object must be exported.
+	if !obj.Exported() {
+		return "", "", ""
+	}
+
+	var recv types.Object // If non-nil, the field or method receiver base.
+
 	switch obj := obj.(type) {
 	case *types.Var:
 		// If the object is a field, and we have an associated selector
 		// composite literal, or struct, we can determine the link.
-		if obj.IsField() {
-			if named, ok := i.enclosing.(*types.Named); ok {
-				rTypeName = named.Obj().Name()
-			}
+		if obj.IsField() && enclosing != nil {
+			recv = enclosing
 		}
 	case *types.Func:
 		typ, ok := obj.Type().(*types.Signature)
 		if !ok {
-			return h, nil
+			// Note: this should never happen. go/types guarantees that the type of
+			// *Funcs are Signatures.
+			//
+			// TODO(rfindley): given a 'debug' mode, we should panic here.
+			return "", "", ""
 		}
 		if r := typ.Recv(); r != nil {
-			switch rtyp := Deref(r.Type()).(type) {
-			case *types.Struct:
-				rTypeName = r.Name()
-			case *types.Named:
+			if rtyp, _ := Deref(r.Type()).(*types.Named); rtyp != nil {
 				// If we have an unexported type, see if the enclosing type is
 				// exported (we may have an interface or struct we can link
 				// to). If not, don't show any link.
 				if !rtyp.Obj().Exported() {
-					if named, ok := i.enclosing.(*types.Named); ok && named.Obj().Exported() {
-						rTypeName = named.Obj().Name()
+					if enclosing != nil {
+						recv = enclosing
 					} else {
-						return h, nil
+						return "", "", ""
 					}
 				} else {
-					rTypeName = rtyp.Obj().Name()
+					recv = rtyp.Obj()
 				}
 			}
 		}
 	}
-	if obj.Pkg() == nil {
-		event.Log(ctx, fmt.Sprintf("nil package for %s", obj))
-		return h, nil
+
+	if recv != nil && !recv.Exported() {
+		return "", "", ""
 	}
-	h.importPath = obj.Pkg().Path()
-	h.LinkPath = h.importPath
-	if mod, version, ok := moduleAtVersion(h.LinkPath, i); ok {
-		h.LinkPath = strings.Replace(h.LinkPath, mod, mod+"@"+version, 1)
+
+	// Either the object or its receiver must be in the package scope.
+	scopeObj := obj
+	if recv != nil {
+		scopeObj = recv
 	}
-	if rTypeName != "" {
-		h.LinkAnchor = fmt.Sprintf("%s.%s", rTypeName, obj.Name())
-		h.symbolName = fmt.Sprintf("(%s.%s).%s", obj.Pkg().Name(), rTypeName, obj.Name())
-		return h, nil
+	if scopeObj.Pkg() == nil || scopeObj.Pkg().Scope().Lookup(scopeObj.Name()) != scopeObj {
+		return "", "", ""
 	}
-	// For most cases, the link is "package/path#symbol".
-	h.LinkAnchor = obj.Name()
-	h.symbolName = fmt.Sprintf("%s.%s", obj.Pkg().Name(), obj.Name())
-	return h, nil
+
+	importPath = obj.Pkg().Path()
+	if recv != nil {
+		anchor = fmt.Sprintf("%s.%s", recv.Name(), obj.Name())
+		name = fmt.Sprintf("(%s.%s).%s", obj.Pkg().Name(), recv.Name(), obj.Name())
+	} else {
+		// For most cases, the link is "package/path#symbol".
+		anchor = obj.Name()
+		name = fmt.Sprintf("%s.%s", obj.Pkg().Name(), obj.Name())
+	}
+	return name, importPath, anchor
 }
 
 func moduleAtVersion(path string, i *IdentifierInfo) (string, string, bool) {
+	// TODO(rfindley): moduleAtVersion should not be responsible for deciding
+	// whether or not the link target supports module version links.
 	if strings.ToLower(i.Snapshot.View().Options().LinkTarget) != "pkg.go.dev" {
 		return "", "", false
 	}
@@ -251,7 +442,7 @@ func moduleAtVersion(path string, i *IdentifierInfo) (string, string, bool) {
 func objectString(obj types.Object, qf types.Qualifier, inferred *types.Signature) string {
 	// If the signature type was inferred, prefer the preferred signature with a
 	// comment showing the generic signature.
-	if sig, _ := obj.Type().(*types.Signature); sig != nil && len(typeparams.ForSignature(sig)) > 0 && inferred != nil {
+	if sig, _ := obj.Type().(*types.Signature); sig != nil && typeparams.ForSignature(sig).Len() > 0 && inferred != nil {
 		obj2 := types.NewFunc(obj.Pos(), obj.Pkg(), obj.Name(), inferred)
 		str := types.ObjectString(obj2, qf)
 		// Try to avoid overly long lines.
@@ -283,11 +474,19 @@ func objectString(obj types.Object, qf types.Qualifier, inferred *types.Signatur
 	return str
 }
 
-// HoverInfo returns a HoverInformation struct for an ast node and its type
-// object. node should be the actual node used in type checking, while fullNode
-// could be a separate node with more complete syntactic information.
-func HoverInfo(ctx context.Context, s Snapshot, pkg Package, obj types.Object, pkgNode ast.Node, fullDecl ast.Decl) (*HoverInformation, error) {
-	var info *HoverInformation
+// FindHoverContext returns a HoverContext struct for an AST node and its
+// declaration object. node should be the actual node used in type checking,
+// while fullNode could be a separate node with more complete syntactic
+// information.
+func FindHoverContext(ctx context.Context, s Snapshot, pkg Package, obj types.Object, pkgNode ast.Node, fullDecl ast.Decl) (*HoverContext, error) {
+	var info *HoverContext
+
+	// Type parameters get their signature from their declaration object.
+	if _, isTypeName := obj.(*types.TypeName); isTypeName {
+		if _, isTypeParam := obj.Type().(*typeparams.TypeParam); isTypeParam {
+			return &HoverContext{signatureSource: obj}, nil
+		}
+	}
 
 	// This is problematic for a number of reasons. We really need to have a more
 	// general mechanism to validate the coherency of AST with type information,
@@ -306,7 +505,7 @@ func HoverInfo(ctx context.Context, s Snapshot, pkg Package, obj types.Object, p
 		// The package declaration.
 		for _, f := range pkg.GetSyntax() {
 			if f.Name == pkgNode {
-				info = &HoverInformation{comment: f.Doc}
+				info = &HoverContext{Comment: f.Doc}
 			}
 		}
 	case *ast.ImportSpec:
@@ -320,12 +519,12 @@ func HoverInfo(ctx context.Context, s Snapshot, pkg Package, obj types.Object, p
 			// so pick the first file that has a doc comment.
 			for _, file := range imp.GetSyntax() {
 				if file.Doc != nil {
-					info = &HoverInformation{source: obj, comment: file.Doc}
+					info = &HoverContext{signatureSource: obj, Comment: file.Doc}
 					break
 				}
 			}
 		}
-		info = &HoverInformation{source: node}
+		info = &HoverContext{signatureSource: node}
 	case *ast.GenDecl:
 		switch obj := obj.(type) {
 		case *types.TypeName, *types.Var, *types.Const, *types.Func:
@@ -337,37 +536,64 @@ func HoverInfo(ctx context.Context, s Snapshot, pkg Package, obj types.Object, p
 			// obj may not have been produced by type checking the AST containing
 			// node, so we need to be careful about using token.Pos.
 			tok := s.FileSet().File(obj.Pos())
-			offset := tok.Offset(obj.Pos())
-			tok2 := s.FileSet().File(node.Pos())
+			offset, err := Offset(tok, obj.Pos())
+			if err != nil {
+				return nil, err
+			}
+
+			// fullTok and fullPos are the *token.File and object position in for the
+			// full AST.
+			fullTok := s.FileSet().File(node.Pos())
+			fullPos, err := Pos(fullTok, offset)
+			if err != nil {
+				return nil, err
+			}
+
 			var spec ast.Spec
 			for _, s := range node.Specs {
 				// Avoid panics by guarding the calls to token.Offset (golang/go#48249).
-				// TODO(rstambler): Investigate this further and adjust if needed.
-				if InRange(tok2, s.Pos()) && InRange(tok2, s.End()) && tok2.Offset(s.Pos()) <= offset && offset <= tok2.Offset(s.End()) {
+				start, err := Offset(fullTok, s.Pos())
+				if err != nil {
+					return nil, err
+				}
+				end, err := Offset(fullTok, s.End())
+				if err != nil {
+					return nil, err
+				}
+				if start <= offset && offset <= end {
 					spec = s
 					break
 				}
 			}
-			var err error
-			info, err = formatGenDecl(node, spec, obj, obj.Type())
+
+			info, err = hoverGenDecl(node, spec, fullPos, obj)
 			if err != nil {
 				return nil, err
 			}
 		}
 	case *ast.TypeSpec:
 		if obj.Parent() == types.Universe {
-			if obj.Name() == "error" {
-				info = &HoverInformation{source: node}
-			} else {
-				info = &HoverInformation{source: node.Name} // comments not needed for builtins
+			if genDecl, ok := fullDecl.(*ast.GenDecl); ok {
+				info = hoverTypeSpec(node, genDecl)
 			}
 		}
 	case *ast.FuncDecl:
 		switch obj.(type) {
 		case *types.Func:
-			info = &HoverInformation{source: obj, comment: node.Doc}
+			info = &HoverContext{signatureSource: obj, Comment: node.Doc}
 		case *types.Builtin:
-			info = &HoverInformation{source: node.Type, comment: node.Doc}
+			info = &HoverContext{Comment: node.Doc}
+			if sig, err := NewBuiltinSignature(ctx, s, obj.Name()); err == nil {
+				info.signatureSource = "func " + sig.name + sig.Format()
+			} else {
+				// Fall back on the object as a signature source.
+
+				// TODO(rfindley): refactor so that we can report bugs from the source
+				// package.
+
+				// debug.Bug(ctx, "invalid builtin hover", "did not find builtin signature: %v", err)
+				info.signatureSource = obj
+			}
 		case *types.Var:
 			// Object is a function param or the field of an anonymous struct
 			// declared with ':='. Skip the first one because only fields
@@ -386,18 +612,13 @@ func HoverInfo(ctx context.Context, s Snapshot, pkg Package, obj types.Object, p
 				if comment.Text() == "" {
 					comment = field.Comment
 				}
-				info = &HoverInformation{source: obj, comment: comment}
+				info = &HoverContext{signatureSource: obj, Comment: comment}
 			}
 		}
 	}
 
 	if info == nil {
-		info = &HoverInformation{source: obj}
-	}
-
-	if info.comment != nil {
-		info.FullDocumentation = info.comment.Text()
-		info.Synopsis = doc.Synopsis(info.FullDocumentation)
+		info = &HoverContext{signatureSource: obj}
 	}
 
 	return info, nil
@@ -421,35 +642,34 @@ func isFunctionParam(obj types.Object, node *ast.FuncDecl) bool {
 	return false
 }
 
-func formatGenDecl(node *ast.GenDecl, spec ast.Spec, obj types.Object, typ types.Type) (*HoverInformation, error) {
-	if _, ok := typ.(*types.Named); ok {
-		switch typ.Underlying().(type) {
-		case *types.Interface, *types.Struct:
-			return formatGenDecl(node, spec, obj, typ.Underlying())
-		}
-	}
+// hoverGenDecl returns hover information an object declared via spec inside
+// of the GenDecl node. obj is the type-checked object corresponding to the
+// declaration, but may have been type-checked using a different AST than the
+// given nodes; fullPos is the position of obj in node's AST.
+func hoverGenDecl(node *ast.GenDecl, spec ast.Spec, fullPos token.Pos, obj types.Object) (*HoverContext, error) {
 	if spec == nil {
-		return nil, errors.Errorf("no spec for node %v at position %v", node, obj.Pos())
+		return nil, errors.Errorf("no spec for node %v at position %v", node, fullPos)
 	}
 
 	// If we have a field or method.
 	switch obj.(type) {
 	case *types.Var, *types.Const, *types.Func:
-		return formatVar(spec, obj, node), nil
+		return hoverVar(spec, fullPos, obj, node), nil
 	}
 	// Handle types.
 	switch spec := spec.(type) {
 	case *ast.TypeSpec:
-		return formatTypeSpec(spec, node), nil
+		return hoverTypeSpec(spec, node), nil
 	case *ast.ValueSpec:
-		return &HoverInformation{source: spec, comment: spec.Doc}, nil
+		return &HoverContext{signatureSource: spec, Comment: spec.Doc}, nil
 	case *ast.ImportSpec:
-		return &HoverInformation{source: spec, comment: spec.Doc}, nil
+		return &HoverContext{signatureSource: spec, Comment: spec.Doc}, nil
 	}
 	return nil, errors.Errorf("unable to format spec %v (%T)", spec, spec)
 }
 
-func formatTypeSpec(spec *ast.TypeSpec, decl *ast.GenDecl) *HoverInformation {
+// TODO(rfindley): rename this function.
+func hoverTypeSpec(spec *ast.TypeSpec, decl *ast.GenDecl) *HoverContext {
 	comment := spec.Doc
 	if comment == nil && decl != nil {
 		comment = decl.Doc
@@ -457,15 +677,13 @@ func formatTypeSpec(spec *ast.TypeSpec, decl *ast.GenDecl) *HoverInformation {
 	if comment == nil {
 		comment = spec.Comment
 	}
-	return &HoverInformation{
-		source:      spec.Type,
-		comment:     comment,
-		typeName:    spec.Name.Name,
-		isTypeAlias: spec.Assign.IsValid(),
+	return &HoverContext{
+		signatureSource: spec,
+		Comment:         comment,
 	}
 }
 
-func formatVar(node ast.Spec, obj types.Object, decl *ast.GenDecl) *HoverInformation {
+func hoverVar(node ast.Spec, fullPos token.Pos, obj types.Object, decl *ast.GenDecl) *HoverContext {
 	var fieldList *ast.FieldList
 	switch spec := node.(type) {
 	case *ast.TypeSpec:
@@ -493,18 +711,18 @@ func formatVar(node ast.Spec, obj types.Object, decl *ast.GenDecl) *HoverInforma
 		// associated values so that we can augment their hover with more information.
 		if _, ok := obj.(*types.Var); ok && spec.Type == nil && len(spec.Values) > 0 {
 			if _, ok := spec.Values[0].(*ast.BasicLit); ok {
-				return &HoverInformation{source: spec, comment: comment}
+				return &HoverContext{signatureSource: spec, Comment: comment}
 			}
 		}
 
-		return &HoverInformation{source: obj, comment: comment}
+		return &HoverContext{signatureSource: obj, Comment: comment}
 	}
 
 	if fieldList != nil {
-		comment := findFieldComment(obj.Pos(), fieldList)
-		return &HoverInformation{source: obj, comment: comment}
+		comment := findFieldComment(fullPos, fieldList)
+		return &HoverContext{signatureSource: obj, Comment: comment}
 	}
-	return &HoverInformation{source: obj, comment: decl.Doc}
+	return &HoverContext{signatureSource: obj, Comment: decl.Doc}
 }
 
 // extractFieldList recursively tries to extract a field list.
@@ -550,11 +768,8 @@ func findFieldComment(pos token.Pos, fieldList *ast.FieldList) *ast.CommentGroup
 	return nil
 }
 
-func FormatHover(h *HoverInformation, options *Options) (string, error) {
-	signature := h.Signature
-	if signature != "" && options.PreferredContentFormat == protocol.Markdown {
-		signature = fmt.Sprintf("```go\n%s\n```", signature)
-	}
+func FormatHover(h *HoverJSON, options *Options) (string, error) {
+	signature := formatSignature(h, options)
 
 	switch options.HoverKind {
 	case SingleLine:
@@ -568,26 +783,50 @@ func FormatHover(h *HoverInformation, options *Options) (string, error) {
 		}
 		return string(b), nil
 	}
+
 	link := formatLink(h, options)
-	switch options.HoverKind {
-	case SynopsisDocumentation:
-		doc := formatDoc(h.Synopsis, options)
-		return formatHover(options, signature, link, doc), nil
-	case FullDocumentation:
-		doc := formatDoc(h.FullDocumentation, options)
-		return formatHover(options, signature, link, doc), nil
+	doc := formatDoc(h, options)
+
+	var b strings.Builder
+	parts := []string{signature, doc, link}
+	for i, el := range parts {
+		if el != "" {
+			b.WriteString(el)
+
+			// Don't write out final newline.
+			if i == len(parts) {
+				continue
+			}
+			// If any elements of the remainder of the list are non-empty,
+			// write a newline.
+			if anyNonEmpty(parts[i+1:]) {
+				if options.PreferredContentFormat == protocol.Markdown {
+					b.WriteString("\n\n")
+				} else {
+					b.WriteRune('\n')
+				}
+			}
+		}
 	}
-	return "", errors.Errorf("no hover for %v", h.source)
+	return b.String(), nil
 }
 
-func formatLink(h *HoverInformation, options *Options) string {
+func formatSignature(h *HoverJSON, options *Options) string {
+	signature := h.Signature
+	if signature != "" && options.PreferredContentFormat == protocol.Markdown {
+		signature = fmt.Sprintf("```go\n%s\n```", signature)
+	}
+	return signature
+}
+
+func formatLink(h *HoverJSON, options *Options) string {
 	if !options.LinksInHover || options.LinkTarget == "" || h.LinkPath == "" {
 		return ""
 	}
 	plainLink := BuildLink(options.LinkTarget, h.LinkPath, h.LinkAnchor)
 	switch options.PreferredContentFormat {
 	case protocol.Markdown:
-		return fmt.Sprintf("[`%s` on %s](%s)", h.symbolName, options.LinkTarget, plainLink)
+		return fmt.Sprintf("[`%s` on %s](%s)", h.SymbolName, options.LinkTarget, plainLink)
 	case protocol.PlainText:
 		return ""
 	default:
@@ -607,37 +846,20 @@ func BuildLink(target, path, anchor string) string {
 	return link + "#" + anchor
 }
 
-func formatDoc(doc string, options *Options) string {
+func formatDoc(h *HoverJSON, options *Options) string {
+	var doc string
+	switch options.HoverKind {
+	case SynopsisDocumentation:
+		doc = h.Synopsis
+	case FullDocumentation:
+		doc = h.FullDocumentation
+	}
 	if options.PreferredContentFormat == protocol.Markdown {
 		return CommentToMarkdown(doc)
 	}
 	return doc
 }
 
-func formatHover(options *Options, x ...string) string {
-	var b strings.Builder
-	for i, el := range x {
-		if el != "" {
-			b.WriteString(el)
-
-			// Don't write out final newline.
-			if i == len(x) {
-				continue
-			}
-			// If any elements of the remainder of the list are non-empty,
-			// write a newline.
-			if anyNonEmpty(x[i+1:]) {
-				if options.PreferredContentFormat == protocol.Markdown {
-					b.WriteString("\n\n")
-				} else {
-					b.WriteRune('\n')
-				}
-			}
-		}
-	}
-	return b.String()
-}
-
 func anyNonEmpty(x []string) bool {
 	for _, el := range x {
 		if el != "" {
