@@ -150,7 +150,6 @@ type symbolCollector struct {
 	matchers   []matcherFunc
 	symbolizer symbolizer
 
-	seen map[span.URI]bool
 	symbolStore
 }
 
@@ -177,9 +176,11 @@ func newSymbolCollector(matcher SymbolMatcher, style SymbolStyle, query string)
 func buildMatcher(matcher SymbolMatcher, query string) matcherFunc {
 	switch matcher {
 	case SymbolFuzzy:
-		return parseQuery(query)
+		return parseQuery(query, newFuzzyMatcher)
 	case SymbolFastFuzzy:
-		return fuzzy.NewSymbolMatcher(query).Match
+		return parseQuery(query, func(query string) matcherFunc {
+			return fuzzy.NewSymbolMatcher(query).Match
+		})
 	case SymbolCaseSensitive:
 		return matchExact(query)
 	case SymbolCaseInsensitive:
@@ -195,6 +196,18 @@ func buildMatcher(matcher SymbolMatcher, query string) matcherFunc {
 	panic(fmt.Errorf("unknown symbol matcher: %v", matcher))
 }
 
+func newFuzzyMatcher(query string) matcherFunc {
+	fm := fuzzy.NewMatcher(query)
+	return func(chunks []string) (int, float64) {
+		score := float64(fm.ScoreChunks(chunks))
+		ranges := fm.MatchedRanges()
+		if len(ranges) > 0 {
+			return ranges[0], score
+		}
+		return -1, score
+	}
+}
+
 // parseQuery parses a field-separated symbol query, extracting the special
 // characters listed below, and returns a matcherFunc corresponding to the AND
 // of all field queries.
@@ -207,7 +220,7 @@ func buildMatcher(matcher SymbolMatcher, query string) matcherFunc {
 // In all three of these special queries, matches are 'smart-cased', meaning
 // they are case sensitive if the symbol query contains any upper-case
 // characters, and case insensitive otherwise.
-func parseQuery(q string) matcherFunc {
+func parseQuery(q string, newMatcher func(string) matcherFunc) matcherFunc {
 	fields := strings.Fields(q)
 	if len(fields) == 0 {
 		return func([]string) (int, float64) { return -1, 0 }
@@ -238,15 +251,7 @@ func parseQuery(q string) matcherFunc {
 				return -1, 0
 			})
 		default:
-			fm := fuzzy.NewMatcher(field)
-			f = func(chunks []string) (int, float64) {
-				score := float64(fm.ScoreChunks(chunks))
-				ranges := fm.MatchedRanges()
-				if len(ranges) > 0 {
-					return ranges[0], score
-				}
-				return -1, score
-			}
+			f = newMatcher(field)
 		}
 		funcs = append(funcs, f)
 	}
